/**
 * api组件
 */
const koaRouter = require('koa-router');
const multer = require('koa-multer');
const send = require('koa-send');

const File = require('./file');
const Moment = require('./moment');
const Log = require('./log');
const _ = require('./lodash');
const Joi = require('./joi');
const ValidateError = require('./errors/validate');
const SecurityError = require('./errors/security');
const WeChatPay = require('./pay/wechat');

const staticPath = `${process.cwd()}/static`;

// 默认允许上传的文件的后缀名
const defaultFileSuffixs = ['jpg', 'jpeg', 'png'];
// 默认上传限制
const defaultLimit = {
  // 文件最大长度 (字节单位)
  fileSize: 1048576, // 1MB = 1024 * 1024Byte
  // 文件最大数量
  files: 5,
};

class Api {
  constructor() {
    this.router = koaRouter();
  }

  getRouter() {
    return this.router;
  }

  /**
   * 验证参数
   * @param ctx
   * @param schema
   */
  validateParameter(ctx, schema) {
    // request validate
    if (!_.isEmpty(schema)) {
      const data = ctx.request.method === 'GET' ? ctx.request.query : ctx.request.body;
      const joiData = {};
      const joiSchema = {};
      /**
       * example: {
         *  name: {
         *    required: true,
         *  },
         *  value: {
         *    type: number,
         *    min: 1,
         *  },
         * }
       */
      _.forIn(schema, (ruleObject, ruleKey) => {
        const type = ruleObject.type || 'string';
        let ruleJoi = Joi[type]();

        // 去除type属性
        ruleObject = _.omit(ruleObject, ['type']);

        // 循环遍历属性，生成ruleJoi
        _.forIn(ruleObject, (value, key) => {
          if (_.isBoolean(value)) {
            if (value) {
              ruleJoi = ruleJoi[key]();
            }
          } else {
            ruleJoi = ruleJoi[key](value);
          }
        });

        // 生成joiData
        joiData[ruleKey] = data[ruleKey];

        // 生成joiSchema
        joiSchema[ruleKey] = ruleJoi;
      });

      const validateResult = Joi.validate(joiData, Joi.object().keys(joiSchema));
      if (validateResult.error !== null) {
        throw new ValidateError(validateResult.error);
      }
    }
  }

  /**
   * 鉴权
   * @param ctx
   * @param needSecurity
   */
  async authenticate(ctx, needSecurity) {
    const needRoles = needSecurity.roles || [];
    const needPermissions = needSecurity.permissions || [];

    // 不需要鉴权
    if (needRoles.length === 0 && needPermissions === 0) {
      return;
    }

    const userSecurity = ctx.state.token.security || {};
    const userRoles = userSecurity.roles || [];
    const userPermissions = userSecurity.permissions || [];

    // 判断角色
    if (needRoles.length > 0) {
      const mergeRoles = _.intersection(needRoles, userRoles);
      if (mergeRoles.length === 0) {
        throw new SecurityError();
      }
    }

    // 判断权限
    if (needPermissions.length > 0) {
      const mergePermissions = _.intersection(needPermissions, userPermissions);
      if (mergePermissions.length === 0) {
        throw new SecurityError();
      }
    }
  }

  /**
   * 统一处理
   * @param opts
   *   - {String} url default ''
   *   - {String} method default 'get'
   *   - {Function} service
   *   - {Object} validate default {}
   *   - {Object} security default {}
   */
  route(opts) {
    const url = opts.url || '';
    const method = opts.method || 'get';
    const service = opts.service;
    const validateSchema = opts.validate || {};
    const needSecurity = opts.security || {};

    if (!service || !_.isFunction(service)) {
      throw new Error('every api must be has service function');
    }

    this.router[method](url, async (ctx) => {
      this.validateParameter(ctx, validateSchema);
      await this.authenticate(ctx, needSecurity);
      const result = await service(ctx);
      ctx.render.success(result);
    });
  }

  get(opts) {
    opts = opts || {};
    opts.method = 'get';
    this.route(opts);
  }

  post(opts) {
    opts = opts || {};
    opts.method = 'post';
    this.route(opts);
  }

  put(opts) {
    opts = opts || {};
    opts.method = 'put';
    this.route(opts);
  }

  del(opts) {
    opts = opts || {};
    opts.method = 'del';
    this.route(opts);
  }

  all(opts) {
    opts = opts || {};
    opts.method = 'all';
    this.route(opts);
  }

  /**
   * 生成文件过滤器
   * @param fileSuffixs
   * @returns {fileFilter}
   */
  generateFileFilter(fileSuffixs) {
    const fileFilter = (req, file, cb) => {
      const fileSuffix = File.suffix(file.originalname);
      if (_.indexOf(fileSuffixs, fileSuffix) === -1) {
        cb(new Error(`${fileSuffix} file do not allow, only allow ${fileSuffixs}`));
      } else {
        cb(null, true);
      }
    };

    return fileFilter;
  }

  /**
   * 生成上传限制器
   * @param fileLimit
   * @returns {*}
   */
  generateFileLimit(fileLimit) {
    if (!fileLimit.fileSize) {
      fileLimit.fileSize = defaultLimit.fileSize;
    }

    if (!fileLimit.files) {
      fileLimit.files = defaultLimit.files;
    }

    return fileLimit;
  }

  /**
   * upload
   * @param opts
   *   - {String} url default ''
   *   - {String} fieldName default 'file'
   *   - {String} filePath default '/uploads'
   *   - {String} fileNameStrategy default ''
   *   - {Number} fileSize default 1
   *   - {Array} fileSuffixs default ['jpg', 'jpeg', 'png']
   *   - {Object} fileLimit default {}
   *   - {Object} validate default {}
   *   - {Object} security default {}
   *   - {Boolean} memory default false
   *   - {Function} service
   */
  upload(opts) {
    opts = opts || {};

    const url = opts.url || '';
    const fieldName = opts.fieldName || 'file';
    const filePath = opts.filePath || '/uploads';
    const fileNameStrategy = opts.fileNameStrategy || '';
    // 上传文件大小
    const fileSize = opts.fileSize || 1;
    // 允许上传的文件的后缀名
    const fileSuffixs = opts.fileSuffixs || defaultFileSuffixs;
    // 上传限制
    const fileLimit = opts.fileLimit || {};
    const validateSchema = opts.validate || {};
    const needSecurity = opts.security || {};
    // 内存模式
    const memory = opts.memory || false;
    const service = opts.service;

    if (!service || !_.isFunction(service)) {
      throw new Error('every upload must be has service function');
    }

    let storage = multer.diskStorage({
      destination: (req, file, cb) => {
        cb(null, `${staticPath}${filePath}`);
      },
      filename: (req, file, cb) => {
        let name = `${file.fieldname}-${Moment().format('YYYY-MM-DD-HH:mm:ss')}-${file.originalname}`;

        if (fileNameStrategy === 'original') {
          name = `${file.originalname}`;
        }

        cb(null, name);
      },
    });

    if (memory) {
      storage = multer.memoryStorage();
    }

    const fileFilter = this.generateFileFilter(fileSuffixs);
    const limits = this.generateFileLimit(fileLimit);

    const uploader = multer({
      storage,
      fileFilter,
      limits,
    });

    this.router.post(url, uploader.array(fieldName, fileSize), async (ctx) => {
      this.validateParameter(ctx, validateSchema);
      await this.authenticate(ctx, needSecurity);
      const result = await service(ctx);
      ctx.render.success(result);
    });
  }

  /**
   * download
   * @param opts
   *   - {String} url default ''
   *   - {String} paramName default 'fileName'
   *   - {Object} security default {}
   */
  download(opts) {
    opts = opts || {};

    const url = opts.url || '';
    const paramName = opts.paramName || 'fileName';
    const needSecurity = opts.security || {};

    this.router.get(url, async (ctx) => {
      const fileName = ctx.request.query[paramName];

      const validateResult = Joi.validate({
        fileName,
      }, Joi.object().keys({
        fileName: Joi.string().required(),
      }));

      if (validateResult.error !== null) {
        throw new ValidateError(validateResult.error);
      }

      await this.authenticate(ctx, needSecurity);

      ctx.attachment(fileName);
      await send(ctx, fileName, {
        root: staticPath,
      });

      Log.info(`render finish ${ctx.request.url} `);
    });
  }

  /**
   * 微信支付、退款、扫码支付回调
   * @param opts
   *   - {String} url default ''
   *   - {String} type default 'pay' pay(支付回调, 默认)、refund(退款回调)、nativePay(扫码支付回调)
   *   - {Function} service
   */
  weChatPayNotify(opts) {
    opts = opts || {};

    const url = opts.url || '';
    const type = opts.type || 'pay';
    const service = opts.service;

    if (!service || !_.isFunction(service)) {
      throw new Error('every weChatPayNotify must be has service function');
    }

    this.router.post(url, WeChatPay.api.middleware(type), async (ctx) => {
      const result = await service(ctx);
      // 回复消息(参数为空回复成功, 传值则为错误消息)
      ctx.reply(result || '');
    });
  }
}

module.exports = (opts) => {
  opts = opts || {};

  // not new router
  const notNew = opts.notNew || false;

  if (notNew) {
    return Api;
  }

  return new Api();
};
