/**
 * 模型组件
 */
const Sequelize = require('sequelize');

const Log = require('./log');
const _ = require('./lodash');
const Moment = require('./moment');
const Uuid = require('./uuid');

let res = {
  database: 'test',
  username: 'root',
  password: 'root',
};

try {
  res = require(`${process.cwd()}/configs/${process.env.NODE_ENV}/model.json`);
} catch (e) {
  Log.warn(`Not found model.json from configs/${process.env.NODE_ENV} folder. Use default value ${JSON.stringify(res)}`);
}

const formatter = 'YYYY-MM-DD HH:mm:ss';

const Op = Sequelize.Op;

const operatorsAliases = {
  $eq: Op.eq,
  $ne: Op.ne,
  $gte: Op.gte,
  $gt: Op.gt,
  $lte: Op.lte,
  $lt: Op.lt,
  $not: Op.not,
  $in: Op.in,
  $notIn: Op.notIn,
  $is: Op.is,
  $like: Op.like,
  $notLike: Op.notLike,
  $iLike: Op.iLike,
  $notILike: Op.notILike,
  $regexp: Op.regexp,
  $notRegexp: Op.notRegexp,
  $iRegexp: Op.iRegexp,
  $notIRegexp: Op.notIRegexp,
  $between: Op.between,
  $notBetween: Op.notBetween,
  $overlap: Op.overlap,
  $contains: Op.contains,
  $contained: Op.contained,
  $adjacent: Op.adjacent,
  $strictLeft: Op.strictLeft,
  $strictRight: Op.strictRight,
  $noExtendRight: Op.noExtendRight,
  $noExtendLeft: Op.noExtendLeft,
  $and: Op.and,
  $or: Op.or,
  $any: Op.any,
  $all: Op.all,
  $values: Op.values,
  $col: Op.col,
};

/**
 * 创建sequelize实例
 */
const sequelize = new Sequelize(res.database, res.username, res.password, {
  host: res.host || 'localhost',
  port: res.port || 3306,
  dialect: res.dialect || 'mysql',
  timezone: res.timezone || '+08:00',
  pool: {
    max: 5,
    min: 0,
    acquire: 30000,
    idle: 10000,
  },
  operatorsAliases,
});

module.exports = {
  sequelize,

  boolean() {
    return Sequelize.BOOLEAN;
  },

  integer() {
    return Sequelize.INTEGER;
  },

  double() {
    return Sequelize.DOUBLE(11, 2);
  },

  enums() {
    return Sequelize.ENUM;
  },

  date() {
    return Sequelize.DATE;
  },

  string(length) {
    return Sequelize.STRING(length);
  },

  char(length) {
    return Sequelize.CHAR(length);
  },

  text(type) {
    return type ? Sequelize.TEXT(type) : Sequelize.TEXT;
  },

  /**
   * 定义模型
   * @param table
   * @param fields
   * @param opts
   * @returns {Model}
   */
  define(table, fields, opts) {
    const bean = {};

    bean.id = {
      type: Sequelize.STRING(50),
      primaryKey: true,
    };

    _.forEach(fields, (value, key) => {
      bean[key] = fields[key];
    });

    opts = opts || {};
    opts.tableName = table;
    opts.timestamps = true;

    opts.hooks = {
      beforeCreate(bean) {
        if (!bean.id) {
          bean.id = Uuid.generate();
        }
      },
      beforeBulkCreate(records) {
        _.forEach(records, (value, key) => {
          if (!records[key].id) {
            records[key].id = Uuid.generate();
          }
        });
      },
    };

    opts.getterMethods = {
      createdAt() {
        const value = this.getDataValue('createdAt');
        if (value) {
          return Moment(value).format(formatter);
        }
        return null;
      },
      updatedAt() {
        const value = this.getDataValue('updatedAt');
        if (value) {
          return Moment(value).format(formatter);
        }
        return null;
      },
    };

    return sequelize.define(table, bean, opts);
  },
};
