const Koa = require('koa');
const cors = require('kcors');
const mount = require('koa-mount');
const serve = require('koa-static');
const favicon = require('koa-favicon');
const bodyParser = require('koa-bodyparser');
const jwt = require('koa-jwt');
const helmet = require('koa-helmet');
const http = require('http');
const onerror = require('koa-onerror');

// 组件
const Log = require('./components/log');
const CoreError = require('./components/errors/core');
const Redis = require('./components/redis');
const Token = require('./components/token');
const _ = require('./components/lodash');
const File = require('./components/file');
const Api = require('./components/api');
const Moment = require('./components/moment');
const Model = require('./components/model');
const Digest = require('./components/digest');
const Captcha = require('./components/captcha');
const Http = require('./components/http');
const Push = require('./components/push');
const Schedule = require('./components/schedule');
const Cheerio = require('./components/cheerio');
const Pinyin = require('./components/pinyin');
const Bagpipe = require('./components/bagpipe');
const Gt = require('./components/gt');
const Random = require('./components/random');
const Math = require('./components/math');
const Excel = require('./components/excel');
const AliSms = require('./components/sms/ali');
const Uuid = require('./components/uuid');
const Socket = require('./components/socket');
const Validator = require('./components/validator');
const Util = require('./components/util');
const WeChat = require('./components/wechat');
const Joi = require('./components/joi');
const AliOss = require('./components/oss/ali');
const WeChatPay = require('./components/pay/wechat');
const WeChatEnterPrise = require('./components/enterprise/wechat');
const Dict = require('./components/dict');

// 中间件
const RateLimitMiddleware = require('./middlewares/rateLimit');
const LogMiddleware = require('./middlewares/log');
const RenderMiddleware = require('./middlewares/render');
const ApiMiddleware = require('./middlewares/api');
const ExtendMiddleware = require('./middlewares/extend');
const NotFoundMiddleware = require('./middlewares/notFound');
const TokenMiddleware = require('./middlewares/token');
const ScheduleMiddleware = require('./middlewares/schedule');

const staticPath = `${process.cwd()}/static`;
const faviconFile = `${process.cwd()}/favicon.ico`;

/**
 * 启动
 * @param {Object} opts
 *   - {Boolean} cors default false
 *   - {Integer} port default '3000'
 *   - {String} tokenSecret default 'token_secret'
 *   - {Integer} expired default 7 days
 *   - {String} apiPrefix default '/api'
 *   - {String} staticPrefix default '/static'
 *   - {Array} middlewares default []
 *   - {Array} apis default []
 *   - {Array} unTokenPaths default []
 *   - {Boolean} logEnable default false
 *   - {String} tokenErrorCode default 'token_error'
 *   - {Boolean} rateLimitDisabled default false
 */
function Boot(opts) {
  opts = opts || {};

  const needCors = opts.cors || false;
  const port = opts.port || 3000;
  const tokenSecret = opts.tokenSecret || 'token_secret';
  const expired = opts.expired || 604800; // 7 days
  const apiPrefix = opts.apiPrefix || '/api';
  const staticPrefix = opts.staticPrefix || '/static';
  const middlewares = opts.middlewares || [];
  const apis = opts.apis || [];
  const unTokenPaths = opts.unTokenPaths || [];
  const logEnable = opts.logEnable || false;
  const tokenErrorCode = opts.tokenErrorCode || 'token_error';
  const rateLimitDisabled = opts.rateLimitDisabled || false;

  // root url do not need token
  unTokenPaths.push('/');

  // token component init
  Token.init(tokenSecret, expired);

  const app = new Koa();

  onerror(app);

  // open cors
  if (needCors) {
    app.use(cors());
  }

  app.use(helmet());
  app.use(mount(staticPrefix, serve(staticPath)));
  app.use(favicon(faviconFile));
  app.use(bodyParser({
    enableTypes: ['json', 'form', 'text'],
    extendTypes: {
      text: ['text/xml', 'application/xml'],
    },
  }));

  if (logEnable) {
    app.use(LogMiddleware.renderStart());
  }

  app.use(RenderMiddleware(logEnable, tokenErrorCode));
  if (!rateLimitDisabled) {
    app.use(RateLimitMiddleware());
  }
  app.use(jwt({ secret: tokenSecret }).unless({ path: unTokenPaths }));
  app.use(TokenMiddleware());
  app.use(ExtendMiddleware(middlewares, app));
  app.use(ScheduleMiddleware());
  app.use(ApiMiddleware(app, apiPrefix, apis));
  app.use(NotFoundMiddleware());

  const server = http.Server(app.callback());

  // socket component init
  Socket.init(server);

  server.listen(port);
  LogMiddleware.appStart(port);
}

module.exports = {
  Boot,
  Log,
  CoreError,
  Redis,
  Token,
  _,
  File,
  Api,
  Moment,
  Model,
  Digest,
  Captcha,
  Http,
  Push,
  Schedule,
  Cheerio,
  Pinyin,
  Bagpipe,
  Gt,
  Random,
  Math,
  Excel,
  AliSms,
  Uuid,
  Socket,
  Validator,
  Util,
  WeChat,
  Joi,
  AliOss,
  WeChatPay,
  WeChatEnterPrise,
  Dict,
};
